import re
from typing import List, Literal, Optional, Tuple

from docstring_parser import parse, DocstringParam
from pydantic import BaseModel


class VarSpec(BaseModel):
    kind: Literal["input", "output"]
    name: str
    description: Optional[str] = None
    datatypes: List[Literal["FILE", "FP32", "FP64", "INT32", "str", "dict"]]
    content_types: Optional[List[str]] = None
    required: Optional[bool] = None


class ParamSpec(BaseModel):
    name: str
    description: Optional[str] = None
    param_types: List[str]


def _get_shape_spec(var_spec: VarSpec):
    var_char = "входной" if var_spec.kind == "input" else "выходной"

    description = f"Размерность значения {var_char} переменной.\n" \
                  "Может быть массивом целых чисел или целым числом.\n" \
                  "Применимо к любым типам данных {var_char}переменной.\n"
    return {
        "description": description,
        "oneOf": [
            {
                "type": "array",
                "items": {
                    "type": "integer"
                },
                "minItems": 1,
                "maxItems": 32
            },
            {
                "type": "integer"
            }
        ]
    }


def _get_output_field_data_spec(var_spec: VarSpec):
    if 'FILE' in var_spec.datatypes:
        description = "Локация, куда следует сохранить выходной файл при datatype == \"FILE\".\n\n" \
                      "Если значение data соответствует файлу, то значение может быть\n" \
                      "сохранено в указанный файл.\n\n" \
                      "Если значение data соответствует файловой группе, то значение может быть \n" \
                      "сохранено в созданный временный файл в заданной файловой группе.\n\n" \
                      "Если значение data не задано, то значение может быть \n" \
                      "сохранено в созданный временный файл в созданной временной файловой группе.\n\n" \
                      "Выражение \"может быть\" означает, что описанный сценарий будет выполнен, \n" \
                      "если инференс функция не подменит переданное ей в атрибуте data значение.\n"
        return {
            "type": "string",
            "description": description
        }
    return None


def _get_input_output_data_spec(var_spec: VarSpec):
    var_char = "входной" if var_spec.kind == "input" else "выходной"
    description = f"Значение {var_char} переменной.\n"
    if 'FILE' in var_spec.datatypes:
        description += "При datatype == \"FILE\" содержит строку с локацией файла.\n"
    if any(dt != 'FILE' for dt in var_spec.datatypes):
        description += "При datatype != \"FILE\" содержит объект как сериализованную строку.\n"
    return {
        "type": "string",
        "description": description
    }


def _get_name_spec(var_spec: VarSpec):
    var_char = "входной" if var_spec.kind == "input" else "выходной"
    return {
        "const": var_spec.name,
        "description": f"Имя {var_char} переменной."
    }


def _get_output_field_datatype_spec(var_spec: VarSpec):
    return {
        "type": "string",
        "description": f"Ожидаемый тип выходной переменной.",
        "enum": var_spec.datatypes
    }


def _get_input_output_datatype_spec(var_spec: VarSpec):
    var_char = "входной" if var_spec.kind == "input" else "выходной"
    return {
        "type": "string",
        "description": f"Тип {var_char} переменной.",
        "enum": var_spec.datatypes
    }


def _get_output_field_content_type_spec(var_spec: VarSpec):
    if 'FILE' in var_spec.datatypes and var_spec.content_types:
        return {
            "type": "string",
            "description": "Ожидаемый тип содержимого файла, \n"
                           "при ожидаемом типе выходной переменной datatype == \"FILE\".\n",
            "enum": var_spec.content_types
        }
    return None


def _get_input_output_content_type_spec(var_spec: VarSpec):
    var_char = "входной" if var_spec.kind == "input" else "выходной"
    if 'FILE' in var_spec.datatypes and var_spec.content_types:
        return {
            "type": "string",
            "description": "Тип содержимого файла, \n"
                           f"при типе {var_char} переменной datatype == \"FILE\".\n",
            "enum": var_spec.content_types
        }
    return None


def _get_base_var_spec(description: str = None):
    openapi_spec = {
        "type": "object",
        "properties": {},
        "required": [
            "name",
            "datatype"
        ]
    }
    if description:
        openapi_spec['description'] = description

    return openapi_spec


def _counstruct_output_field_spec(var_spec: VarSpec):
    openapi_spec = _get_base_var_spec(var_spec.description)
    openapi_spec['properties']['name'] = _get_name_spec(var_spec)
    openapi_spec['properties']['datatype'] = _get_output_field_datatype_spec(var_spec)
    data_spec = _get_output_field_data_spec(var_spec)
    if data_spec:
        openapi_spec['properties']['data'] = data_spec
    content_type = _get_output_field_content_type_spec(var_spec)
    if content_type:
        openapi_spec['properties']['content_type'] = content_type
    return openapi_spec


def _construct_input_output_var_spec(var_spec: VarSpec):
    openapi_spec = _get_base_var_spec(var_spec.description)
    openapi_spec['properties']['name'] = _get_name_spec(var_spec)
    openapi_spec['properties']['datatype'] = _get_input_output_datatype_spec(var_spec)
    openapi_spec['properties']['data'] = _get_input_output_data_spec(var_spec)
    openapi_spec['required'].append("data")
    shape = _get_shape_spec(var_spec)
    openapi_spec['properties']['shape'] = shape
    openapi_spec['required'].append("shape")
    if var_spec.content_types:
        content_type = _get_input_output_content_type_spec(var_spec)
        if content_type:
            openapi_spec['properties']['content_type'] = content_type
    return openapi_spec


def _construct_model_param_spec(param_spec: ParamSpec):
    openapi_spec = {
        "type": "array",
        "prefixItems": [
            {
                "const": param_spec.name
            },
            {
                "type": "object",
                "additionalProperties": {
                    "type": param_spec.param_types
                }
            }
        ],
        "minItems": 2,
        "maxItems": 2
    }
    if param_spec.description:
        openapi_spec['description'] = param_spec.description

    return openapi_spec


def _parse_docstring_var_name(name_str: str):
    parts = name_str.split('.')
    if len(parts) != 2:
        return None, None
    if parts[0] not in {'inputs', 'outputs', 'model_parameters'}:
        return None, None
    parts[0] = parts[0][:-1]  # remove trailing 's'
    return parts


def _parse_docstring_var_type(type_str: str, supported_datatypes, supported_content_types):
    required = True
    if type_str.endswith('optional'):
        required = False
        type_str = type_str.rstrip('optional').rstrip(' ').rstrip(',')

    parts = re.split(r'or\s+', type_str)
    parts[0].strip()

    parts = [re.sub(r'[\s"\']*$', '', p) for p in parts]
    parts = [re.sub(r'^[\s"\']*', '', p) for p in parts]
    parts = set(parts)
    datatypes = parts & set(supported_datatypes)
    content_types = parts & set(supported_content_types)
    if len(content_types) == 0:
        content_types = None
    if content_types and ('FILE' not in datatypes):
        datatypes.add('FILE')
    return datatypes, content_types, required


def _parse_docstring_param_type(type_str: str, supported_param_types):
    if not type_str:
        return ['string']

    if type_str.endswith('optional'):
        type_str = type_str.rstrip('optional').rstrip(' ').rstrip(',')

    parts = re.split(r'or\s+', type_str)
    parts[0].strip()

    parts = [re.sub(r'[\s"\']*$', '', p) for p in parts]
    parts = [re.sub(r'^[\s"\']*', '', p) for p in parts]
    parts = set(parts)
    param_types = parts & set(supported_param_types)
    if not param_types:
        return ['string']
    return param_types


def _create_spec_for_var(var_name, var_kind, param: DocstringParam,
                         supported_datatypes, supported_content_types) -> VarSpec:
    if not param.type_name:
        raise ValueError(f'param {param.arg_name} type not specified')
    dict_var = {
        'name': var_name,
        'kind': var_kind,
        'description': param.description
    }

    datatypes, content_types, required = _parse_docstring_var_type(param.type_name,
                                                                   supported_datatypes,
                                                                   supported_content_types)
    dict_var['datatypes'] = datatypes
    if content_types:
        dict_var['content_types'] = content_types
    if dict_var['kind'] == 'input':
        dict_var['required'] = required

    spec_var = VarSpec(**dict_var)
    return spec_var


def _create_spec_for_param(var_name: str, param: DocstringParam, supported_param_types) -> ParamSpec:
    dict_param = {
        'name': var_name,
        'description': param.description,
        'param_types': _parse_docstring_param_type(param.type_name, supported_param_types)
    }

    spec_param = ParamSpec(**dict_param)
    return spec_param


def _create_spec_from_docstring(docstring: str, supported_datatypes, supported_content_types, supported_param_types):
    docstring_struct = parse(docstring)
    processed_params = set()
    spec_vars = []
    spec_model_params = []
    for param in docstring_struct.params:

        var_key = param.arg_name
        if var_key in processed_params:
            continue
        var_kind, var_name = _parse_docstring_var_name(param.arg_name)
        if not var_name:
            continue
        if var_kind == 'input' or var_kind == 'output':
            spec_var = _create_spec_for_var(var_name, var_kind, param, supported_datatypes, supported_content_types)
            spec_vars.append(spec_var)
        if var_kind == 'model_parameter':
            spec_param = _create_spec_for_param(var_name, param, supported_param_types)
            spec_model_params.append(spec_param)
        processed_params.add(var_key)

    required_input_vars = [spec_var for spec_var in spec_vars
                           if spec_var.required and spec_var.kind == 'input']
    optional_input_vars = [spec_var for spec_var in spec_vars
                           if not spec_var.required and spec_var.kind == 'input']
    output_vars = [spec_var for spec_var in spec_vars
                   if spec_var.kind == 'output']

    inputs = [_construct_input_output_var_spec(v) for v in required_input_vars]
    inputs_min_items = len(inputs)
    inputs += [_construct_input_output_var_spec(v) for v in optional_input_vars]
    inputs_max_items = len(inputs)
    outputs = [_construct_input_output_var_spec(v) for v in output_vars]
    output_fields = [_counstruct_output_field_spec(v) for v in output_vars]
    output_max_items = len(outputs)
    model_parameters = [_construct_model_param_spec(p) for p in spec_model_params]

    inputs_spec, output_fields_spec, outputs_spec, model_parameters_spec = (None, ) * 4

    if outputs:
        outputs_spec = {
            "description": "Выходные переменные, возвращаются как результат из инференс функции.\n",
            "type": "array",
            "prefixItems": outputs,
            "maxItems": output_max_items
        }

    if output_fields:
        output_fields_spec = {
            "description": "Список выходных переменных, которые инференс функция должна вернуть.",
            "type": "array",
            "prefixItems": output_fields,
            "maxItems": output_max_items
        }

    if inputs:
        inputs_spec = {
            "type": "array",
            "description": "Входные переменные, передаются как аргумент в инференс функцию.\n",
            "prefixItems": inputs,
            "minItems": inputs_min_items,
            "maxItems": inputs_max_items
        }

    if model_parameters:
        model_parameters_spec = {
            "type": "array",
            "description": "Параметры, передаются как аргумент в инференс функцию.\n"
                           "Каждый параметр представляется кортежем из двух элементов (имя, значение).\n",
            # это нужно для параметров модели, которые не специфицированы явно в схеме
            "items": {
                "type": "array",
                "prefixItems": [
                    {"type": "string"},
                    {"type": "object"}
                ],
                "minItems": 2,
                "maxItems": 2
            },
            "prefixItems": model_parameters
        }

    return inputs_spec, output_fields_spec, outputs_spec, model_parameters_spec


def create_args_spec(spec_source, datatypes, content_types, param_types) -> Tuple[dict, dict, dict, dict]:
    try:
        if spec_source.__doc__:
            return _create_spec_from_docstring(spec_source.__doc__, datatypes, content_types, param_types)
    except Exception as exc:
        raise ValueError("Error on spec create") from exc
    raise ValueError("Only object with docstring as source is supported now")
