import base64
import logging
import os
import posixpath
import urllib.parse
import uuid
from importlib import import_module
from io import BytesIO
from string import ascii_uppercase, ascii_lowercase
from typing import List, Literal, Tuple, Optional, cast, Union, get_args

import requests
from fastapi import FastAPI, Request, Depends
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse, PlainTextResponse
from pydantic import BaseModel, Extra, ConfigDict
from typing_extensions import Annotated

from unip.mlcmp.config import FILES_S3_MODE, UNIP_MLCMP_APP_NAME, FILES_S3_BOX_NAME, UNIP_FILES_API, \
    FILES_S3_COPY_INFERENCE_FILES_DIR, INFERENCE_MODEL_PATH, INFERENCE_FUNCTION
from unip.mlcmp.inference_function_protocol import InferenceOutput, FILE_DATA_TYPE, InferenceInput, \
    InferenceOutputField
from unip.mlcmp.logging import configure_logging
from unip.mlcmp.openapi import create_openapi_spec
from unip.mlcmp.request_id import RequestIdMiddleware
from unip.mlcmp.license import get_license


UNIP_FILES_USER_PASSWORD = '**'
DATATYPES = Literal["FILE", "FP32", "FP64", "INT32", "dict", "str"]
CONTENT_TYPES = Literal["text/csv", "application/json"]
PARAM_TYPES = ["integer", "number", "string", "boolean"]

configure_logging()

logger = logging.getLogger(__name__)

app = FastAPI(openapi_url=None)
app.add_middleware(RequestIdMiddleware)


def load_inference_func():
    index = INFERENCE_FUNCTION.rindex('.')
    model_module = import_module(INFERENCE_FUNCTION[:index])
    func_name = INFERENCE_FUNCTION[index + 1:]
    inf_func = getattr(model_module, func_name)
    return inf_func


inference_func = load_inference_func()


def construct_opeanpi_spec():
    try:
        spec = create_openapi_spec(inference_function=inference_func,
                                   datatypes=get_args(DATATYPES),
                                   content_types=get_args(CONTENT_TYPES),
                                   param_types=PARAM_TYPES)
        return spec
    except Exception as exc:
        logger.exception('OpenAPI specification generation error, '
                         'specification not available', exc_info=exc)
        return None


openapi_spec = construct_opeanpi_spec()

service_license = get_license()

cached_model = None


class Input(BaseModel, extra=Extra.allow):
    name: str
    data: object
    datatype: DATATYPES
    content_type: Optional[str] = None
    shape: Union[int, List[int]]


class OutputField(BaseModel):
    name: str
    data: Optional[str] = None
    datatype: DATATYPES
    content_type: Optional[str] = None


class PredictRequest(BaseModel):
    model_parameters: Optional[List[Tuple[object, object]]] = None
    inputs: List[Input]
    output_fields: List[OutputField]

    model_config = ConfigDict(
        protected_namespaces=()
    )


class Output(BaseModel):
    name: str
    data: object  # presigned_get_url for file
    datatype: DATATYPES
    content_type: Optional[str] = None
    shape: Union[int, List[int]]


class PredictResponse(BaseModel):
    outputs: List[Output]


class ModelVersionResponse(BaseModel):
    openapi_spec: Optional[dict]
    license: Optional[str]
    model_version: str = '1'

    model_config = ConfigDict(
        protected_namespaces=()
    )


class UnipMlcmpException(Exception):
    pass


class InputValdationException(UnipMlcmpException):
    pass


class AccessDeniedException(UnipMlcmpException):
    pass


class InferenceException(UnipMlcmpException):
    pass


class InputFileNotFoundException(UnipMlcmpException):
    pass


class LicenseFileNotFoundException(UnipMlcmpException):
    pass


@app.exception_handler(InputValdationException)
async def input_validation_exception_handler(request: Request, exc: InputValdationException):
    logger.error('Cannot process request with given data', exc_info=exc)
    return JSONResponse(
        status_code=400,
        content={
            'errors': [{
                'type': 'Bad request',
                'message': str(exc)}]
        })


@app.exception_handler(AccessDeniedException)
async def access_denied_exception_handler(request: Request, exc: AccessDeniedException):
    logger.error('Cannot process request with given data', exc_info=exc)
    return JSONResponse(
        status_code=403,
        content={
            'errors': [{
                'type': 'Access denied',
                'message': str(exc)}]
        })


@app.exception_handler(InferenceException)
async def inference_exception_handler(request: Request, exc: InferenceException):
    logger.error('Server error when running inference', exc_info=exc)
    # if sys.version_info >= (3, 10):
    #     exc_str = ''.join(traceback.format_exception(exc))
    # else:
    #     exc_str = ''.join(traceback.format_exception(None, exc, exc.__traceback__))
    return JSONResponse(
        status_code=422,
        content={
            'errors': [{
                'type': 'Unprocessable entity',
                'message': f'Server error when running inference function'}]
        })


@app.exception_handler(InputFileNotFoundException)
async def input_file_not_found_exception_handler(request: Request, exc: InputFileNotFoundException):
    logger.error('Server error when running inference', exc_info=exc)
    return JSONResponse(
        status_code=404,
        content={
            'errors': [{
                'type': 'Not found',
                'message': f'Input file given location not found. {exc}'}]
        })


@app.exception_handler(LicenseFileNotFoundException)
async def license_file_not_found_exception_handler(request: Request, exc: LicenseFileNotFoundException):
    return JSONResponse(
        status_code=404,
        content={
            'errors': [{
                'type': 'Not found',
                'message': f'License file doesnt exist.'}]
        })


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    # todo
    # list(set([])) нужно, потому что pydantic дублирует сообщения об ошибках (нужно разобраться, почему)
    formatted_validation_errors = list(set([f'Validation error for input {err["input"]}. '
                                            f'Location: {err["loc"]}; type: {err["type"]}; '
                                            f'message: {err["msg"]}'
                                            for err in exc.errors()]))
    for e in formatted_validation_errors:
        logger.error(f'Error in input data: {e}')
    errors = [{
        'type': 'Bad request',
        'message': e}
        for e in formatted_validation_errors]

    return JSONResponse(
        status_code=400,
        content={
            'errors': errors
        })


def _create_object_inference_inputs(request_inputs: List[Input]) -> List[InferenceInput]:
    inputs = []
    for request_input in request_inputs:
        if request_input.datatype == FILE_DATA_TYPE:
            continue
        input_ = InferenceInput(name=request_input.name,
                                data=request_input.data,
                                datatype=request_input.datatype,
                                content_type=request_input.content_type,
                                shape=request_input.shape)
        for extra_field in request_input.model_extra:
            input_[extra_field] = getattr(request_input, extra_field)
        inputs.append(input_)
    return inputs


def _get_file_from_api(user_id: str, file_group: str, file_name: str):
    app_name = UNIP_MLCMP_APP_NAME
    box_name = FILES_S3_BOX_NAME
    location = posixpath.join(file_group, file_name)
    get_file_uri = f'/{app_name}/files/{box_name}/{location}'
    payload = {'user_id': user_id}
    get_file_url = urllib.parse.urljoin(UNIP_FILES_API, get_file_uri)

    response = requests.get(get_file_url, json=payload, auth=(user_id, UNIP_FILES_USER_PASSWORD))

    if response.status_code != 200:
        if response.status_code == 404:
            raise InputFileNotFoundException(f'File \'{location}\' not found')
        logging.error(f'Error while requesting \'{get_file_url}\'')
        logging.error(f'Response status code: {response.status_code}')
        logging.error(f'Response content: {response.content}')
        raise UnipMlcmpException(f'Cant get file \'{location}\' by get url')

    response_data = response.json()
    return response_data


def _get_presigned_get_url(user_id: str, file_group: str, file_name: str):
    response_data = _get_file_from_api(user_id, file_group, file_name)
    return response_data['presigned_get_url']


def _get_local_file_path():
    # здесь можно использовать file_group/file_name, но это не обязательно,
    # проще использовать случайное имя
    tmp_name = uuid.uuid4().hex
    local_file_path = os.path.join(FILES_S3_COPY_INFERENCE_FILES_DIR, tmp_name)
    local_dir = os.path.join(FILES_S3_COPY_INFERENCE_FILES_DIR)
    os.makedirs(local_dir, exist_ok=True)
    return local_file_path


def _get_object_to_file(presigned_get_url, file_group, file_name):
    local_file_path = _get_local_file_path()
    location = posixpath.join(file_group, file_name)
    response = requests.get(presigned_get_url, allow_redirects=True)

    if response.status_code != 200:
        if response.status_code == 404:
            raise InputFileNotFoundException(f'Content of input file \'{location}\' doesnt exist;'
                                             f'given get url: \'{presigned_get_url}\'')
        logging.error(f'Error while getting input file {location} content by url \'{presigned_get_url}\'')
        logging.error(f'Response status code: {response.status_code}')
        logging.error(f'Response content: {response.content}')
        raise UnipMlcmpException(f'Cant get content of input file \'{location}\' '
                                 f'by get url \'{presigned_get_url}\'')

    obj = BytesIO(response.content)

    with open(local_file_path, 'wb') as f:
        f.write(obj.getbuffer())

    return local_file_path


def _get_input_file_by_copying(user_id, file_group, file_name):
    presigned_get_url = _get_presigned_get_url(user_id, file_group, file_name)
    local_file_path = _get_object_to_file(presigned_get_url, file_group, file_name)
    return local_file_path


def _get_input_file(user_id, file_group, file_name):
    if FILES_S3_MODE == 'COPY':
        return _get_input_file_by_copying(user_id, file_group, file_name)
    raise UnipMlcmpException(f'Unsupported FILES_S3_MODE {FILES_S3_MODE}')


def _parse_location(location: str, allow_empty=False, allow_empty_file=False) -> Tuple[Optional[str], Optional[str]]:

    file_group_name = None
    file_name = None

    if location in {'/', '', None}:
        if allow_empty:
            return None, None
        else:
            raise InputValdationException(f'Location not specified')

    loc_unquoted = urllib.parse.unquote(location)
    loc_norm = posixpath.normpath(loc_unquoted)
    if loc_norm.startswith('/'):
        loc_norm = loc_norm[1:]
    left_part, right_segment = posixpath.split(loc_norm)
    if left_part == '':
        if not allow_empty_file:
            raise InputValdationException(f'Location \'{location}\' doesnt contain file name')
        else:
            file_group_name = right_segment
    elif loc_unquoted.endswith(posixpath.sep):
        file_group_name = loc_norm
    else:
        file_group_name = left_part
        file_name = right_segment

    return file_group_name, file_name


def _create_file_inference_inputs(user_id: str, request_inputs: List[Input]) -> List[InferenceInput]:
    inputs = []
    for request_input in request_inputs:
        if request_input.datatype != FILE_DATA_TYPE:
            continue

        file_group, file_name = _parse_location(cast(str, request_input.data))

        local_file_path = _get_input_file(user_id, file_group=file_group,
                                          file_name=file_name)

        input_ = InferenceInput(name=request_input.name,
                                data=local_file_path,
                                datatype=request_input.datatype,
                                content_type=request_input.content_type,
                                shape=request_input.shape)

        inputs.append(input_)
    return inputs


def _create_inference_inputs(user_id, request_inputs: List[Input]) -> List[InferenceInput]:
    inputs = []
    inputs += _create_object_inference_inputs(request_inputs)
    file_inputs = _create_file_inference_inputs(user_id, request_inputs)
    inputs += file_inputs
    return inputs


def _create_inference_output_fields(request_output_fields: List[OutputField]) \
        -> List[InferenceOutputField]:
    output_fields = []
    for request_output_field in request_output_fields:
        data = _get_local_file_path() if request_output_field.datatype == "FILE" else None

        output_field = InferenceOutputField(name=request_output_field.name,
                                            datatype=request_output_field.datatype,
                                            content_type=request_output_field.content_type,
                                            data=data)
        output_fields.append(output_field)
    return output_fields


def _run_inference(inputs: List[InferenceInput],
                   output_fields: List[InferenceOutputField],
                   parameters: List[Tuple[object, object]]) -> List[InferenceOutput]:
    global cached_model
    parameters = parameters if parameters else []
    try:
        inference_outputs, model = inference_func(model_key=INFERENCE_MODEL_PATH,
                                                  inputs=[i.model_dump() for i in inputs],
                                                  output_fields=[of.model_dump() for of in output_fields],
                                                  parameters=parameters,
                                                  model=cached_model)
        inference_outputs = [InferenceOutput.model_validate(io) for io in inference_outputs]
    except Exception as exc:
        raise InferenceException from exc
    # временное решение, нужно сохранять модель в быстрый реестр\кэш
    # todo: fix
    if model:
        cached_model = model
    return inference_outputs


def _get_or_create_output_file_urls(user_id: str, file_group: Optional[str], file_name: Optional[str]) -> Tuple[str, str]:
    app_name = UNIP_MLCMP_APP_NAME
    box_name = FILES_S3_BOX_NAME
    # предполагает, что file_group и file_name не начинаются и не оканчиваются символом '/'
    file_group = file_group if file_group is not None else ''
    file_name = file_name if file_name is not None else ''
    location = f'{file_group}/{file_name}'
    file_uri = f'/{app_name}/files/{box_name}/{location}'
    file_op_url = urllib.parse.urljoin(UNIP_FILES_API, file_uri)
    body = {'user_id': user_id}

    res_data = None
    error_msg = None

    if file_group:
        if file_name:
            # если передан путь к файлу, то нужно получить подписанную ссылку, для этого нужно вызвать PUT
            res = requests.put(file_op_url, auth=(user_id, UNIP_FILES_USER_PASSWORD))
            if res.status_code == 200:
                res_data = res.json()
            else:
                error_msg = f'Cant get presigned_put_url for output file; requested PUT: "{file_op_url}"'
        else:
            # если передан только путь к файловой группе, нужно создать новый файл в файловой группе
            file_op_url = file_op_url + '/'
            res = requests.post(file_op_url, auth=(user_id, UNIP_FILES_USER_PASSWORD))
            if res.status_code == 200 or res.status_code == 201:
                res_data = res.json()
            else:
                error_msg = f'Cant create file for output file, given file_group; requested POST: "{file_op_url}"'
    else:
        # если передана пустая локация, нужно создать новый файл в новой файловой группе
        res = requests.post(file_op_url, auth=(user_id, UNIP_FILES_USER_PASSWORD))
        if res.status_code == 200 or res.status_code == 201:
            res_data = res.json()
        else:
            error_msg = f'Cant create file url for output file; requested url: "{file_op_url}"'

    if res_data:
        get_url = res_data['presigned_get_url']
        put_url = res_data['presigned_put_url']
        return get_url, put_url
    else:
        logging.error(error_msg)
        logging.error(f'Response status code: {res.status_code}')
        logging.error(f'Response content: {res.content}')
        raise UnipMlcmpException(error_msg)


def _put_object_from_file(file_group, file_name, presigned_put_url, local_file_path) -> None:
    with open(local_file_path, 'rb') as f:
        bts = f.read()
        obj = BytesIO(bts)

    headers = {'Content-Type': 'application/json', 'Content-Length': str(len(bts))}
    res = requests.put(presigned_put_url, data=obj, headers=headers)

    if res.status_code != 201 and res.status_code != 200:
        logging.error(f'Error while putting file by url \'{presigned_put_url}\'')
        logging.error(f'Response status code: {res.status_code}')
        logging.error(f'Response content: {res.content}')
        raise UnipMlcmpException(f'Cant put content of output file \'{file_name}\' of file group \'{file_group}\'')


def _put_output_file_by_copying(user_id: str, file_group: str, file_name: str, local_file_path: str):
    presigned_get_url, presigned_put_url = _get_or_create_output_file_urls(user_id, file_group, file_name)
    _put_object_from_file(file_group, file_name, presigned_put_url, local_file_path)
    return presigned_get_url


def _put_output_file(user_id: str, file_group: str, file_name: str, local_file_path: str):
    if FILES_S3_MODE == 'COPY':
        return _put_output_file_by_copying(user_id, file_group, file_name, local_file_path)
    raise UnipMlcmpException(f'Unsupported FILES_S3_MODE {FILES_S3_MODE}')


def _create_file_outputs(user_id: str, output_fields: List[OutputField],
                         inference_outputs: List[InferenceOutput]) -> List[Output]:
    outputs = []
    output_fields_by_name: dict[str, OutputField] = {o.name: o for o in output_fields}
    for inference_output in inference_outputs:
        if inference_output.datatype != FILE_DATA_TYPE or not inference_output.data:
            continue

        file_group, file_name = _parse_location(output_fields_by_name[inference_output.name].data,
                                                allow_empty=True, allow_empty_file=True)
        output_data = cast(str, inference_output.data)

        presigned_get_url = _put_output_file(user_id, file_group, file_name, output_data)

        output = Output(name=inference_output.name,
                        data=presigned_get_url,
                        datatype=inference_output.datatype,
                        content_type=inference_output.content_type,
                        shape=inference_output.shape)
        outputs.append(output)
    return outputs


def _create_object_outputs(inference_outputs: List[InferenceOutput]) -> List[Output]:
    outputs = []
    for inference_output in inference_outputs:
        if inference_output.datatype == FILE_DATA_TYPE:
            continue
        outputs.append(Output(**inference_output.model_dump()))
    return outputs


def _create_outputs(user_id: str, output_fields: List[OutputField],
                    inference_outputs: List[InferenceOutput]) -> List[Output]:
    file_outputs = _create_file_outputs(user_id, output_fields, inference_outputs)
    outputs = file_outputs + _create_object_outputs(inference_outputs)
    return outputs


# https://datatracker.ietf.org/doc/html/rfc3986#section-2.3
ALLOWED_USER_NAME_SYMBOLS = set(ascii_uppercase) | set(ascii_lowercase) | set('0123456789') | {'-', '.', '_', '~'}


def _get_user_id(request: Request):
    if 'Authorization' not in request.headers:
        raise AccessDeniedException(f'User identity not passed')
    authorization = request.headers['Authorization']
    parts = authorization.split(' ')
    if parts[0] != 'Basic' or len(parts) < 2:
        raise AccessDeniedException(f'Only Basic Authorization schema is supported now')
    try:
        credentials = base64.b64decode(parts[1]).decode(encoding='utf-8')
        sep_index = credentials.index(':')
    except ValueError:
        raise AccessDeniedException(f'Invalid Basic Authorization credentials format')
    except Exception:
        raise AccessDeniedException(f'Invalid Basic Authorization credentials')
    user_name = credentials[:sep_index]
    if any([s not in ALLOWED_USER_NAME_SYMBOLS for s in user_name]):
        raise AccessDeniedException(f'User name symbols not allowed')
    return user_name


@app.post('/predict', response_model=PredictResponse)
async def predict(predict_request: PredictRequest, user_id: Annotated[str, Depends(_get_user_id)]):
    inference_inputs = _create_inference_inputs(user_id, predict_request.inputs)
    inference_output_fields = _create_inference_output_fields(predict_request.output_fields)
    inference_outputs = _run_inference(inference_inputs,
                                       inference_output_fields,
                                       predict_request.model_parameters)
    outputs = _create_outputs(user_id, predict_request.output_fields,
                              inference_outputs)
    pred_resp = PredictResponse(outputs=outputs)
    return pred_resp


@app.get('/modelversion', response_model=ModelVersionResponse, dependencies=[Depends(_get_user_id)])
async def modelversion():
    return ModelVersionResponse(openapi_spec=openapi_spec,
                                license=service_license.location)


@app.get('/modelversion/license', response_class=PlainTextResponse, dependencies=[Depends(_get_user_id)])
async def local_license():
    if service_license.content:
        return service_license.content
    else:
        raise LicenseFileNotFoundException
