import os
import sys

from unip.api_spec import create_args_spec
import json


def create_openapi_spec(inference_function, datatypes, content_types, param_types) -> dict:
    base_dir = os.path.dirname(os.path.realpath(__file__))
    json_path = os.path.join(base_dir, 'openapi-base.json')

    with open(json_path) as f:
        spec_dict = json.load(f)

    inputs_spec, output_fields_spec, outputs_spec, params_spec = create_args_spec(inference_function,
                                                                                  datatypes,
                                                                                  content_types,
                                                                                  param_types)

    schemas = spec_dict['components']['schemas']
    req_properties = schemas['PredictionRequest']['properties']
    if inputs_spec:
        req_properties['inputs'] = inputs_spec
    if output_fields_spec:
        req_properties['output_fields'] = output_fields_spec
    if params_spec:
        req_properties['model_parameters'] = params_spec
    if outputs_spec:
        resp_properties = schemas['PredictionResponse']['properties']
        resp_properties['outputs'] = outputs_spec

    return spec_dict


if __name__ == '__main__':

    from typing import Any

    def example_inference_func(parameters: list,
                               inputs: list,
                               output_fields: set,
                               model_key: str,
                               model: Any = None) -> (list, Any):
        """
        :param inputs.p: Параметр прогнозирования
        :type inputs.p: FP32, optional
        :param inputs.X: Матрица экземпляров для классификации
        :type inputs.X: FP64 or FP32 or FILE or text/csv
        :param outputs.predict: Результат прогнозирования
        :type outputs.predict: INT32 or 'FILE' or text/csv or application/json
        :param outputs.scores: Скоры прогнозирования
        :type outputs.scores: FP32 or FP64 or UnknownType or text/csv
        :param model_parameters.mode: Режим модели
        :type model_parameters.mode: number or integer of float
        :type model_parameters.mode: string, optional
        """
        pass


    def import_inference_func(func_name):
        index = func_name.rindex('.')
        from importlib import import_module
        func_module = import_module(func_name[:index])
        func_name = func_name[index + 1:]
        func = getattr(func_module, func_name)
        return func


    if len(sys.argv) != 2 and len(sys.argv) != 3:
        raise SyntaxError('At least output path must be passed; inference function full name can be passed to')
    if len(sys.argv) == 3:
        inference_func = import_inference_func(sys.argv[2])
    else:
        inference_func = example_inference_func

    spec = create_openapi_spec(inference_func,
                               ['FILE', 'FP32', 'FP64', 'dict', 'str', 'INT32'],
                               ['text/csv', 'application/json'],
                               ['boolean', 'integer', 'number', 'string'])

    with open(sys.argv[1], 'w', encoding='utf-8') as f:
        json.dump(spec, f, ensure_ascii=False, indent=2)
