import logging
import uuid
from contextvars import ContextVar
from typing import Optional

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from fastapi.responses import JSONResponse

request_id: ContextVar[Optional[str]] = ContextVar('request_id', default=None)


# См.
# https://www.starlette.io/middleware/#basehttpmiddleware
# https://github.com/snok/asgi-correlation-id/tree/main


def create_request_id():
    return str(uuid.uuid4())


def is_valid_request_id(value):
    try:
        uuid.UUID(value, version=4)
    except ValueError:
        return False
    return True


class RequestIdMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):

        if 'X-Request-Id' in request.headers:
            req_id = request.headers['X-Request-Id']
            if not is_valid_request_id(req_id):
                logging.getLogger(__name__).error(f'X-Request-Id must be a valid UUID4 string, given: {req_id}')
                return JSONResponse(
                    status_code=400,
                    content={
                           'errors': [{
                               'type': 'Bad request',
                               'message': f'X-Request-Id must be a valid UUID4 string'}]
                       })
        else:
            req_id = create_request_id()
        request_id.set(req_id)
        logging.getLogger(__name__).info(f'Before request {request.method} {request.url}')

        response = await call_next(request)

        if 'X-Request-Id' not in response.headers:
            if request_id.get() is not None:
                response.headers['X-Request-Id'] = request_id.get()
        logging.getLogger(__name__).info(f'After request {request.method} {request.url}')

        return response

